# This file is part of Pantry, command-line nutrient analysis.
# 
# Version 21, released Monday, September 17, 2007.
# 
# Copyright 2007 Omari Norman.
# 
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the
# Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall
# be included in all copies or substantial portions of the
# Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
# OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# 
# Pantry's "built-in" XML validator is a modified version of
# xmlproc.  xmlproc is Copyright 1998-2000 by Lars Marius
# Garshol, Oslo, Norway.
# 
# All Rights Reserved
# 
# Permission to use, copy, modify, and distribute this
# software and its documentation for any purpose and without
# fee is hereby granted, provided that the above copyright
# notice appear in all copies and that modified copies are
# clearly marked as such.
# 
# LARS MARIUS GARSHOL DISCLAIMS ALL WARRANTIES WITH REGARD TO
# THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL LARS MARIUS
# GARSHOL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
# DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
# USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

""" Main Pantry command-line interface.
    
It is meant to be called using a trivial Python script that
may be installed in the user's PATH. """

from __future__ import division

def main(argv):
    """ argv is the entire argv of the calling script.

    Thus argv must include sys.argv[0] as well as the rest of the argv. """
    from optparse import OptionParser
    import errno
    import os
    import sys
    import textwrap
    
    from pantry.lib import cabinets
    from pantry.lib import foods
    from pantry.lib import io
    from pantry.lib import uniconv
    from pantry.lib import version
    import config
    import reportPrinter

    prog = os.path.basename(argv[0])
    usage = 'usage: %prog [options] FILE...'

    parser = OptionParser(prog = prog, usage = usage)

    # search options
    parser.add_option('-i', '--ignore-case', action='store_const',
            const=cabinets.IGNORECASE,
            dest='ignoreCase', 
            help="make searches case insensitive")
    parser.add_option('-x', '--exact-match', action='store_const',
            const=cabinets.EXACTMATCH,
            dest='exactMatch',
            help='traits must exactly match search term specified')
    parser.add_option("-n", "--name", action="store",
            type="string", dest="name", 
            metavar='regexp',
            help='select foods with matching name')
    parser.add_option('-g', '--group', action="store",
            type='string', dest='group', 
            metavar='regexp',
            help='select foods with matching group')
    parser.add_option("-d", "--date", action="store",
            type="string", dest='date', 
            metavar='regexp',
            help='select foods with matching date')
    parser.add_option('-m', '--meal', action='store',
            type='string', dest='meal', 
            metavar='regexp',
            help='select foods with matching meal')
    parser.add_option('-u', '--unit', action='store',
            type='string', dest='unit', 
            metavar='regexp',
            help='select foods with matching unit')
    parser.add_option('-q', '--qty', action='store',
            type='string', dest='qty', 
            metavar='number',
            help='select foods with matching qty')
    parser.add_option('-c', '--comment', action='store',
            type='string', dest='comment',
            metavar='regexp',
            help='select foods with matching comment')
    parser.add_option('-o', '--order', action='store',
            type='string', dest='order',
            metavar='regexp',
            help='select foods with matching order')
    parser.add_option('-p', '--print', action='store',
            type='string', dest='report', metavar = 'REPORT',
            help='print REPORT')
    parser.add_option('-l', '--nutrient-list', action='store',
            type='string', dest='nutList', metavar='NUTRIENT-LIST',
            help='use NUTRIENT-LIST when printing report (default facts)')
    parser.add_option('-a', '--add', action='append',
            type='string', dest='addFiles', metavar='FILE',
            help='add results to FILE')
    parser.add_option('-s', '--sort', action='store', type='string',
            dest='sort', metavar='traits',
            help='sort foods in report by these traits')
    parser.add_option('--force-valid', action='store_const',
            const = cabinets.BIGVALID, dest='force',
            help='validate all XML files, even if above 300 kilobytes')
    parser.add_option('--skip-valid', action='store_const',
            const = cabinets.SKIPVALID, dest = 'skip',
            help = 'do not validate any XML files')
    
    # change options
    parser.add_option('-N', '--c-name', action='store',
            type='string', dest='c_name', metavar='string',
            help='change name to string')
    parser.add_option('-G', '--c-group', action='store',
            type='string', dest='c_group', metavar='string',
            help='change group to string')
    parser.add_option('-Q', '--c-qty', action='store',
            type='string', dest='c_qty', metavar='number',
            help='change quantity to number')
    parser.add_option('-U', '--c-unit', action='store',
            type='string', dest='c_unit', metavar='regexp',
            help='change unit to one matching regexp')
    parser.add_option('-D', '--c-date', action='store',
            type='string', dest='c_date', metavar='string',
            help='change date to string')
    parser.add_option('-M', '--c-meal', action='store',
            type='string', dest='c_meal', metavar='string',
            help='change meal to string')
    parser.add_option('-C', '--c-comment', action='store',
            type='string', dest='c_comment', metavar='string',
            help='change comment to string')
    parser.add_option('-O', '--c-order', action='store',
            type='string', dest='c_order', metavar='string',
            help='change order to string')
    parser.add_option('-A', '--auto-order', action='store_true',
            dest='autoOrder', 
            help='automatically increment order trait')
    parser.add_option('--by-nut', action='store',
            type='string', dest='c_by_nut', metavar='nut-name qty',
            nargs=2,
            help='set quantity of food so amount of nut-name = qty')
    parser.add_option('-K', '--by-cal', action='store',
            type='string', dest='c_by_cal', metavar='calories',
            help='set quantity of food so calories equals calories')
    parser.add_option('--edit', action='store_true',
            dest='edit', help='Edit foods in place') 
    parser.add_option('--delete', action='store_true', 
            dest='delete', help='delete matching foods')
    parser.add_option('-R', '--refuse', action='store_true',
            dest='useRefuse', 
            help='reduce qty by percent refuse')

    # dump options
    parser.add_option('--dump', action='store', type='string',
            dest='dump', metavar='what-to-dump',
            help='print internal Pantry variables (details in man page)')
    parser.add_option('-v', '--version', action='store_const',
            const='version', dest='dump',
            help='Print version and copyright information and exit')

    parser.set_defaults(ignoreCase=0, exactMatch=0,
            addFiles = [], sort='', autoOrder = False,
            force = 0, skip = 0,)

    (options, posArgs) = parser.parse_args(argv[1:])

    # convert to unicode
    options = uniconv.convertOptions(options, io.ENCODING)
    posArgs = uniconv.convertList(posArgs, io.ENCODING)

    stderr = io.puni(sys.stderr)
    errors = io.ErrorPrinter(stderr, prog, io.WARNING)
    stdout = io.puni(sys.stdout)
    searchOpts = options.ignoreCase | options.exactMatch

    try:
        configVars = config.ConfigVars()
    except config.Error, e:
        errors(u'Problem loading Pantry RC file:', e, 
                type = io.ERROR)

    if options.c_by_cal:
        options.c_by_nut = [u'Calories', options.c_by_cal]

    if options.dump:
        for dumpName, whatToPrint in (('version', version.COPYRIGHT),
            ('pantryDTD', textwrap.dedent(cabinets.PANTRYDTD)),
            ('rcDTD', textwrap.dedent(config.DTD)),
            ('config', str(configVars))):
            if options.dump == dumpName:
                stdout(whatToPrint + '\n')
                sys.exit()
        errors(u'unrecognized argument for dump option.',
                type = io.ERROR)

    if len(posArgs) < 1:
        errors('you did not specify any files to search.')
    if options.ignoreCase and options.exactMatch:
        errors('--ignore-case has no effect when using --exact-match')

    # do we need nutrients?
    needNuts = False
    if options.report:
        if options.nutList:
            nutList = configVars.nutLists[options.nutList]
        else:
            nutList = configVars.nutLists['facts']
        reporter = reportPrinter.Reports(options.report,
                prog, sys.stdout, sys.stderr, options.sort,
                configVars.sortOrders, nutList, configVars.localReports)
        needNuts = reporter.needNuts
    if options.edit or options.addFiles or options.c_by_nut:
        needNuts = True
    results = cabinets.Temp()

    for filename in posArgs:
        # search files
        try:
            target = cabinets.getCabinet(filename, 
                    options = options.skip | options.force)
        except cabinets.CabinetOpenError, e:
            errors('could not open file %s' % filename, e)
            continue

        searchDict = dict([(trait, getattr(options, trait))
                    for trait in ('name', 'group', 'date', 'meal',
                        'qty', 'unit', 'comment', 'order')
                    if getattr(options, trait) != None])
        keyList = target.search(searchDict, searchOpts)

        for key in keyList:
            # change foods; fill results instance
            newFood = target.getFood(key, needNuts)
            newTraits = [(trait, getattr(options, 'c_' + trait))
                    for trait in ('name', 'group', 'date', 'meal',
                        'qty', 'comment', 'order')
                    if getattr(options, 'c_' + trait) != None]
            try:
                newFood.setTraits(newTraits)
            except foods.SetTraitError, e:
                errors('food %s will not be copied into the buffer.' %
                        newFood.traits['name'], e)
                continue

            if options.c_unit != None:
                try:
                    newFood.setUnit(options.c_unit, searchOpts)
                except foods.SetUnitError, e:
                    errors('food %s will not be copied into the buffer.' %
                            newFood.traits['name'], e)
                    continue

            if options.c_by_nut:
                newQty = newFood.getQtyByNut(options.c_by_nut[0],
                        options.c_by_nut[1], searchOpts)
                newFood.traits['qty'] = newQty

            if options.useRefuse:
                newQty = newFood.qtyMinusRefuse
                newFood.traits['qty'] = newQty

            # add food to buffer
            mutex, narf = _addFood(newFood, results, False)
            if mutex:
                errors(u'Food %s' % newFood.traits['name']
                        + u' had duplicate in buffer;'
                        + u' added "(Copy %u)"' % mutex 
                        + u' to its comment trait')

            # edit or delete food if necessary
            if options.edit:
                target.addFood(newFood, key)

            if options.delete:
                del target[key]
            
            if options.edit or options.delete:
                try:
                    target.save()
                except cabinets.ReadOnlyError:
                    errors(u'file %s could not be opened' % filename
                            + ' for writing; no foods in it are edited'
                            + ' or deleted')

        del target

    if options.report:
        reporter.printReports(results)

    for addFile in options.addFiles:
        try:
            saveTarget = cabinets.getCabinet(addFile, 
                    options = options.force | options.skip | cabinets.CREATE)
        except cabinets.CabinetOpenError, e:
            errors(u'could not open for --add file %s' % addFile, e)
            continue

        warnIncrement = False
        for food in results.values():
            mutex, noIncrement = _addFood(food, saveTarget, options.autoOrder)
            if mutex:
                errors(u'duplicate of food %s' % food.traits['name']
                        + u' when adding to file %s;' % addFile
                        + u' added "(Copy %u)" to Comment trait' % mutex)
            if noIncrement and options.autoOrder:
                warnIncrement = True

        if warnIncrement:
            errors(u'--auto-order has no effect when adding foods '
                    u' to file %s; it is a Txt file.' % addFile)

        try:
            saveTarget.save()
        except cabinets.ReadOnlyError:
            errors(u'file %s could not be opened for writing;' % addFile
                    + ' no foods added to it')

def _addFood(food, cabinet, incrementOrder):
    """ Adds a food to a cabinet.

    Catches TypeErrors (which arise when attempting to use
    incrementOrder on Txt cabinets) or DuplicateFoodErrors (and
    automatically mutates foods when they arise).

    Returns 2-tuple. First element is mutex. Second element is True
    if incrementOrder has no effect; False otherwise. """

    from pantry.lib import cabinets
    added = False
    mutex = 0
    noIncrement = False
    while not added:
        try:
            cabinet.addFood(food = food, oldKey = None,
                    incrementOrder = incrementOrder)
        except TypeError:
            # too many arguments; this is a Txt cabinet
            noIncrement = True
            try:
                cabinet.addFood(food, None)
            except cabinets.DuplicateFoodError:
                if mutex == 0:
                    food = food.copy()
                food, mutex = food.mutant
            else:
                added = True
        except cabinets.DuplicateFoodError:
            if mutex == 0:
                food = food.copy()
            food, mutex = food.mutant
        else:
            added = True
    return mutex, noIncrement
