# This file is part of Pantry, command-line nutrient analysis.
# 
# Version 21, released Monday, September 17, 2007.
# 
# Copyright 2007 Omari Norman.
# 
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the
# Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall
# be included in all copies or substantial portions of the
# Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
# OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# 
# Pantry's "built-in" XML validator is a modified version of
# xmlproc.  xmlproc is Copyright 1998-2000 by Lars Marius
# Garshol, Oslo, Norway.
# 
# All Rights Reserved
# 
# Permission to use, copy, modify, and distribute this
# software and its documentation for any purpose and without
# fee is hereby granted, provided that the above copyright
# notice appear in all copies and that modified copies are
# clearly marked as such.
# 
# LARS MARIUS GARSHOL DISCLAIMS ALL WARRANTIES WITH REGARD TO
# THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL LARS MARIUS
# GARSHOL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
# DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
# USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import division

class Reports(object):
    """ Print out food reports. """
    
    def __init__(self, reportsWanted, prog, output, errors, sortFields, 
            sortOrders, nutList, localReportDir):
        from pantry.lib import foods
        from pantry.lib import io
        import reports
        import sys
        import os
        import errno

        reportsWanted = reportsWanted.split('-')
        self.sortOrders = sortOrders
        self.nutList = nutList
        allReporters = reports.allReports.copy() 

        self.err = io.ErrorPrinter(io.puni(errors), prog, io.WARNING)
        if localReportDir:
            files = sorted(os.listdir(localReportDir))
            files = [os.path.join(localReportDir, filename)
                    for filename in files]
            for filename in files:
                execdict = {}
                execfile(filename, execdict)
                allReporters.update(execdict['allReports'])
        
        classesToPrint = []
        for reportName in reportsWanted:
            possRpts = [ rptClass for rptName, rptClass 
                    in allReporters.items()
                    if rptName.startswith(reportName)]
            if len(possRpts) != 1:
                self.err('report name %s does not exist or is ambiguous.' %
                        reportName)
                continue
            classesToPrint.append(possRpts[0])

        self.reportsToPrint = [rptClass(prog, output, errors)
                for rptClass in classesToPrint]

        self.sortList = []
        for char in sortFields:
            for traitName in foods.TRAITSBYNAME:
                found = False
                if traitName.startswith(char):
                    self.sortList.append(traitName)
                    found = True
                    break
                elif traitName.startswith(char.lower()):
                    self.sortList.append('-' + traitName)
                    found = True
                    break
            if not found:
                self.err("sort trait %s not recognized" % char)

        self.needNuts = False
        for rptInstance in self.reportsToPrint:
            try:
                needNuts = rptInstance.needNuts
            except AttributeError:
                needNuts = False
            if needNuts:
                self.needNuts = True
                return

    def printReports(self, cabinet):
        from pantry.lib import foods
        from pantry.lib import cabinets
        import inspect
        goalReporter = cabinets.GoalReporter(self.nutList, cabinet)

        foodReports=[]
        summaryReports=[]
        for rptInstance in self.reportsToPrint:
            if 'food' not in inspect.getargspec(rptInstance.printReport)[0]:
                summaryReports.append(rptInstance)
            else:
                foodReports.append(rptInstance)

        if foodReports:
            if self.sortList:
                foodsToPrint = sorted(cabinet.values(), 
                        foods.CmpFoods(self.sortList, self.sortOrders))
            else:
                foodsToPrint = cabinet.values()

            for food in foodsToPrint:
                for rptInstance in foodReports:
                    argDict = dict([(arg, locals()[arg]) for arg
                        in inspect.getargspec(rptInstance.printReport)[0]
                        if arg != 'self'])
                    rptInstance.printReport(**argDict)
        
        for rptInstance in summaryReports:
            argDict = dict([(arg, locals()[arg]) for arg
                in inspect.getargspec(rptInstance.printReport)[0]
                if arg != 'self'])
            rptInstance.printReport(**argDict)
