# This file is part of Pantry, command-line nutrient analysis.
# 
# Version 21, released Monday, September 17, 2007.
# 
# Copyright 2007 Omari Norman.
# 
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the
# Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall
# be included in all copies or substantial portions of the
# Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
# OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# 
# Pantry's "built-in" XML validator is a modified version of
# xmlproc.  xmlproc is Copyright 1998-2000 by Lars Marius
# Garshol, Oslo, Norway.
# 
# All Rights Reserved
# 
# Permission to use, copy, modify, and distribute this
# software and its documentation for any purpose and without
# fee is hereby granted, provided that the above copyright
# notice appear in all copies and that modified copies are
# clearly marked as such.
# 
# LARS MARIUS GARSHOL DISCLAIMS ALL WARRANTIES WITH REGARD TO
# THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL LARS MARIUS
# GARSHOL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
# DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
# USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from __future__ import division
allReports = {}
def _getReportBase():
    class ReportBase:
        def __init__(self, prog, stdout, stderr):
            from pantry.lib import io
            self.stdout = stdout
            self.stderr = stderr
            self.prog = prog
            self.output = io.puni(stdout)
            self.errors = io.ErrorPrinter(io.puni(stderr), prog, io.WARNING)
    return ReportBase

### NAMES REPORTER
def _namesPrinter(self, food):
    self.output(food.traits['name'] + '\n')

_reporter = _getReportBase()
_reporter.needNuts = False
_reporter.printReport = _namesPrinter
allReports['names'] = _reporter

### YIELD REPORTER
def _yieldPrinter(self, food):
    try:
        if food.rTraits['yieldDesc']:
            self.output(food.rTraits['yieldDesc'] + ' ')
    except AttributeError:
        self.errors('food %s is not a recipe.' % food.traits['name'])
        return
    self.output('(%.0fg)\n' % food.totalGrams)

_reporter = _getReportBase()
_reporter.needNuts = False
_reporter.printReport = _yieldPrinter
allReports['yield'] = _reporter

### INGREDIENTS REPORTER
def _ingredientsPrinter(self, food):
    if not hasattr(food, 'ingredients'):
        self.errors('food %s is not a recipe.' % food.traits['name'])
        return
    for ingr in food.ingredients:
        grams = ingr.grams
        oz = ingr.grams / 28.35
        self.output('%s %s (%.3foz, %.0fg) %s' % (ingr.traits['qty'],
                ingr.traits['unit'], oz, grams, ingr.traits['name']))
        if ingr.traits['comment']:
            self.output(' (%s)' % ingr.traits['comment'])
        self.output('\n')

_reporter = _getReportBase()
_reporter.needNuts = False
_reporter.printReport = _ingredientsPrinter
allReports['ingredients'] = _reporter

### DIRECTIONS REPORTER
def _directionsPrinter(self, food):
    import re
    if not hasattr(food, 'ingredients'):
        self.errors('food %s is not a recipe.' % food.traits['name'])
        return
    directions = food.rTraits['directions']
    directions = re.sub(r'<p[^>]*>', '', directions)
    directions = re.sub(r'\s+', ' ', directions)
    directions = re.sub(r'</p>', '\n', directions)
    self.output(directions)

_reporter = _getReportBase()
_reporter.needNuts = False
_reporter.printReport = _directionsPrinter
allReports['directions'] = _reporter

### RECIPE REPORTER
def _recipePrinter(self, food):
    if not hasattr(food, 'rTraits'):
        self.errors('food %s is not a recipe.' % food.traits['name'])
        return
    
    for report in ('names', 'yield', 'blank', 'ingredients', 'blank',
            'directions'):
        reporter = allReports[report](self.prog, self.stdout, self.stderr)
        reporter.printReport(food)

_reporter = _getReportBase()
_reporter.needNuts = False
_reporter.name = 'makes'
_reporter.printReport = _recipePrinter
allReports['recipe'] = _reporter

### PASTE REPORTER
def _pastePrinter(self, food):
    OPTEXACT='-x'
    OPTNAME='-n'
    OPTCHANGEUNIT='-U'

    name = food.traits['name']
    name = "'%s'" % name.replace("'", "'\\''")
    availUnits = food.units
    availUnits = [ "'%s'" % unit.replace("'", "'\\''")
            for unit in availUnits ]

    for availUnit in availUnits:
        self.output(' '.join([OPTEXACT, OPTNAME, name, 
            OPTCHANGEUNIT, availUnit, '\n']))

_reporter = _getReportBase()
_reporter.needNuts = False
_reporter.printReport = _pastePrinter
allReports['paste'] = _reporter

### TRAITS REPORTER
def _traitsPrinter(self, food):
    from pantry.lib import uniconv

    traits = food.traits

    self.output(traits['name'] + '\n')
    if traits['group']:
        self.output('Group: %s\n' % traits['group'])

    atStartOfLine = True
    for traitName in ['date', 'meal', 'order']:
        if traits[traitName]:
            self.output('%s: %s ' % 
                    (traitName.capitalize(), traits[traitName]))
            atStartOfLine = False
    if not atStartOfLine:
        self.output('\n',)

    if traits['comment']:
        self.output('Comment: %s\n' % traits['comment'])

    if uniconv.convertUni(traits['pctRefuse']):
        self.output(('Refuse: %s percent %s\n' %
                (traits['pctRefuse'], traits['refDesc'])))

    self.output(('%s %s (%.0fg)\n' 
            % (traits['qty'], traits['unit'], food.grams)))

_reporter = _getReportBase()
_reporter.needNuts = False
_reporter.printReport = _traitsPrinter
allReports['traits'] = _reporter

### MEASURES REPORTER
def _measuresPrinter(self, food):
    units = food.units
    for unit, grams in units.items():
        self.output('   %s (%.0fg)\n' % (unit, grams))

_reporter = _getReportBase()
_reporter.needNuts = False
_reporter.printReport = _measuresPrinter
allReports['measures'] = _reporter

### UNITS REPORTER
def _unitsPrinter(self, food):
    from pantry.lib import foods
    units = set(food.units).difference(foods.ABSOLUTES)
    for unit in units:
        self.output('   %s\n' % unit)

_reporter = _getReportBase()
_reporter.needNuts = False
_reporter.printReport = _unitsPrinter
allReports['units'] = _reporter

### DEBUG REPORTER
def _debugPrinter(self, food):
    self.output(str(food) + '\n')

_reporter = _getReportBase()
_reporter.needNuts = True
_reporter.printReport = _debugPrinter
allReports['debug'] = _reporter

### LIST REPORTER
def _listPrinter(self, goalReporter):
    from pantry.lib import foods

    format = "%-25s %-14s %-6s\n"
    headerLabels = ('Nutrient Name', 'Units', 'Goal')
    headerString = format % headerLabels
    self.output(headerString)
    self.output("-" * len(headerString) + '\n')

    for nutName, goal in goalReporter.goals:
        nutUnits = foods.NUTSBYNAME[nutName].units
        self.output(format % (nutName, nutUnits, goal))

_reporter = _getReportBase()
_reporter.needNuts = False
_reporter.printReport = _listPrinter
allReports['list'] = _reporter

def _nutReport(self, food, goalReporter):
    """ Prints nutrient report for food.

    If food is None, prints sum nutrient report.
    """

    headerFormat = "%-25s %-14s %-6s"
    headerLabels = ['Nutrient', 'Amount', '%G' ]
    if food:
        headerFormat += ' %-6s'
        headerLabels.append('%TOT')
    else:
        headerFormat = 'SUM:\n%s' % headerFormat
    headerFormat += '\n'
    headerString = headerFormat % tuple(headerLabels)
    self.output(headerString)
    self.output("-" * len(headerString) + '\n')
    lineFormat = "%-25s %-4s %-10s %-6s"
    if food:
        lineFormat += ' %-6s\n'
        data = goalReporter[food.hash]
    else:
        lineFormat += '\n'
        data = goalReporter.totals

    for line in data:
        toPrint = [line['nutID'][0], line['nutAmt'], 
                line['nutID'][1], line['pctGoal']]
        if food:
            toPrint.append(line['pctTotal'])

        self.output(lineFormat % tuple(toPrint))

### NUTS REPORTER
def _nutsPrinter(self, food, goalReporter):
    _nutReport(self, food, goalReporter)

_reporter = _getReportBase()
_reporter.needNuts = True
_reporter.printReport = _nutsPrinter
allReports['nuts'] = _reporter

### SUM REPORTER
def _sumPrinter(self, goalReporter):
    _nutReport(self, None, goalReporter)

_reporter = _getReportBase()
_reporter.needNuts = True
_reporter.printReport = _sumPrinter
allReports['sum'] = _reporter

### BLANK REPORTER
def _blankPrinter(self, food):
    self.output('\n')

_reporter = _getReportBase()
_reporter.needNuts = False
_reporter.printReport = _blankPrinter
allReports['blank'] = _reporter

### GROUPS REPORTER
def _groupsPrinter(self, cabinet):
    groupDict = cabinet.groups
    self.output('  %-35s %7s\n' % ('Group', 'No. of Foods'))
    self.output('-' * 60 + '\n')
    for group in sorted(groupDict):
        self.output('  %-35s %7u\n' % (group, len(groupDict[group])))
    self.output('-' * 60 + '\n')
    self.output('  %-35s %7u foods total\n' % ('', len(cabinet)))
    self.output('  %-35s %7u groups total\n' % ('', len(groupDict)))

_reporter = _getReportBase()
_reporter.needNuts = False
_reporter.printReport = _groupsPrinter
allReports['groups'] = _reporter
