# This file is part of Pantry, command-line nutrient analysis.
# 
# Version 21, released Monday, September 17, 2007.
# 
# Copyright 2007 Omari Norman.
# 
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the
# Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall
# be included in all copies or substantial portions of the
# Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS
# OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# 
# Pantry's "built-in" XML validator is a modified version of
# xmlproc.  xmlproc is Copyright 1998-2000 by Lars Marius
# Garshol, Oslo, Norway.
# 
# All Rights Reserved
# 
# Permission to use, copy, modify, and distribute this
# software and its documentation for any purpose and without
# fee is hereby granted, provided that the above copyright
# notice appear in all copies and that modified copies are
# clearly marked as such.
# 
# LARS MARIUS GARSHOL DISCLAIMS ALL WARRANTIES WITH REGARD TO
# THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL LARS MARIUS
# GARSHOL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
# DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
# USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

""" Food objects, the most basic Pantry object.

Foods are the most basic object in Pantry. Cabinets hold multiple
Food objects. There are two kinds of foods: Food instances, and
Recipe instances. A Food has its own nutrient information. A Recipe
gets its nutrient information from its ingredient Foods (which may be
Food instances or Recipe instances.)

The first place to look for documentation is the docstrings for the
Recipe and Food objects.
"""

from __future__ import division
import UserDict

class _Nuts(UserDict.DictMixin):
    """ Contains several nutrients.
        
    A dictionary; the keys are nutrientNo,
    and the values are the amount of the nutrient (float). """

    def __init__(self, parent, nuts = None):
        """ Set up new Nuts instance.

        To make empty Nuts instance, leave nuts = None.
        You may also use a mapping instance or a sequence of
        2-tuples, as you can with ordinary dictionaries. """

        self.nutdict = {}
        self.parent = parent
        try:
            nuts = nuts.items()
        except AttributeError:
            pass

        self.setitems(nuts)

    def __getitem__(self, key):
        """ Given key, returns nutrient value. 
        
        Key is a sequence. First value is the name of the nutrient;
        second value is the units. Units is optional (currently, including
        them has no effect. """

        nutNo = NUTSBYNAME[key[0]].no
        return self.nutdict[nutNo] * self.parent.grams / 100 

    def setitems(self, items):
        """ Adds nutrients to this instance.

        items: list or tuple with nuts to add. 
        The first item may be a nutrient number of the nutrient to add.
        In this instance, the nutrient is added, *without* converting
        it to nutsPer100g of food. In this case it is assumed that the
        nutrient has already been converted to nutsPer100g.

        Or, the first item may be a tuple. In this case, the first
        item of the tuple is the name of the nutrient to add. The
        second item is the units. If units are percent, the nutrient
        amount is adjusted accordingly; otherwise, the units is
        checked for correctness. In either case, nutrients are
        converted to nutsPer100g.
        
        Thus, the following are permissible as the first item:
            203
            ('Protein', '%')
            ('Protein', 'g')

        The following are NOT permissible as the first item:
            (203,)
            'Protein'

        The second item is what you want to set the nutrient equal to.
        If first item is an integer, second item must be an integer or a float.
        If first item is a tuple, second item must be a string or Unicode. """

        import uniconv

        grams = self.parent.grams
        for key, value in items:
            if key not in NUTSBYNO:
                try:
                    value = uniconv.convertUni(value)
                except uniconv.UnconvertableError:
                    raise NutValueError(key, value)
                if value is None:
                    raise NutValueError(key, value)
            
                try:
                    nutNo = NUTSBYNAME[key[0]].no
                except KeyError:
                    raise NutIDError(key, value)

                nutUnits = key[1]

                if nutUnits == '%':
                    dvValue = NUTSBYNO[nutNo].dv
                    if dvValue == None:
                        raise NoPercentageError(key, value)
                    value = value / 100 * dvValue
                elif nutUnits != NUTSBYNO[nutNo].units:
                    raise NutUnitMismatchError(key, value, 
                            NUTSBYNO[nutNo].units)

                value = value / grams * 100
                key = nutNo
            self.nutdict[key] = value

    def __setitem__(self, key, value):
        self.setitems(((key, value),))

    def __delitem__(self, key):
        """ Key is a sequence, (nutName, nutUnits).

        Currently the units has no effect though. """

        nutNo = NUTSBYNAME[key[0]].no
        del self.nutdict[nutNo]
    
    def keys(self):
        return [ (NUTSBYNO[nutNo].name, NUTSBYNO[nutNo].units) 
                for nutNo in self.nutdict ]

class _Traits(UserDict.DictMixin):
    """Contains a food's traits.

    From the outside this looks like a dictionary, with the keys
    being the trait names, and the values being the trait values. 
    On the inside, the trait names are stored as integers to save
    space. """

    def __init__(self, parent, initial = None):
        """ Set up new Traits instance. """

        self.parent = parent

        # initialize traits 
        self.traitdict = dict([(traitNo, u'') for traitNo
            in TRAITSBYNO])
        del self.traitdict[TRAITSBYNAME['qty']]
        del self.traitdict[TRAITSBYNAME['unit']]

        try:
            initial = initial.items()
        except AttributeError:
            pass

        self.setitems(initial)

        try:
            self['qty']
        except KeyError:
            raise NoQtyError()
        try:
            self['unit']
        except KeyError:
            raise NoUnitError()

    def setitems(self, items):
        """ Set traits in items.

        Items is a mapping. Keys may be integers corresponding
        to those in TRAITSBYNO, or they may be strings
        corresponding to those in TRAITSBYNAME. Values
        are always Unicodes, even for qty and pctRefuse. """

        import uniconv

        for key, value in items:
            if key not in TRAITSBYNO:
                if key in FLOATTRAITS:
                    try:
                        uniconv.convertUni(value)
                    except uniconv.UnconvertableError:
                        raise TraitValueError(key, value)
                if key == 'unit' and value not in self.parent.units:
                    raise NoAvailUnitError(key, value)
                key = TRAITSBYNAME[key]
            self.traitdict[key] = value

            # update parent grams if necessary
            if key == TRAITSBYNAME['unit'] or key == TRAITSBYNAME['qty']:
                try:
                    qtyStr = self['qty']
                except KeyError:
                    pass
                else:
                    try:
                        qtyNo = uniconv.convertUni(qtyStr)
                    except uniconv.UnconvertableError:
                        raise TraitValueError(key, value)

                    try:
                        gramsPerUnit = self.parent.units[self['unit']]
                    except KeyError:
                        pass
                    else:
                        self.parent._grams = (qtyNo * gramsPerUnit)

    def __setitem__(self, key, value):
        self.setitems(((key, value),))

    def __getitem__(self, key):
        traitNo = TRAITSBYNAME[key]
        return self.traitdict[traitNo]

    def __delitem__(self, key):
        try:
            traitNo = TRAITSBYNAME[key]
        except KeyError:
            traitNo = key
        del self.traitdict[traitNo]

    def keys(self):
        return [ traitName for traitName, traitNo
                in TRAITSBYNAME.items()
                if traitNo in self.traitdict ]

class _RTraits(UserDict.DictMixin):
    """ Holds traits for a recipe. """

    def __init__(self, parent, initial):
        self.parent = parent
        self.traitdict = dict([(traitNo, u'') for traitNo
            in RTRAITSBYNO])
        try:
            initial = initial.items()
        except AttributeError:
            pass
        self.setitems(initial)

    def __getitem__(self, key):
        rTraitNo = RTRAITSBYNAME[key]
        return self.traitdict[rTraitNo]

    def __setitem__(self, key, value):
        self.setitems(((key, value),))

    def __delitem__(self, key):
        try:
            rTraitNo = RTRAITSBYNAME[key]
        except KeyError:
            rTraitNo = key
        del self.traitdict[traitNo]

    def keys(self):
        return [rTraitName for rTraitName, rTraitNo
                in RTRAITSBYNAME.items()
                if rTraitNo in self.traitdict]

    def setitems(self, items):
        import uniconv

        for key, value in items:
            if key not in RTRAITSBYNO:
                if key in RFLOATTRAITS:
                    try:
                        uniconv.convertUni(value)
                    except uniconv.UnconvertableError:
                        raise TraitValueError(key, value)
                key = RTRAITSBYNAME[key]
            self.traitdict[key] = value

class _ArbUnits(UserDict.DictMixin):
    """ Contains a food's arbitrary units. """
    def __init__(self, parent, initial = None):
        self.arbUnitDict = {}
        self.parent = parent
        try:
            initial = initial.items()
        except AttributeError:
            pass

        self.setitems(initial)

    def setitems(self, items):
        import uniconv

        for key, value in items:
            try:
                value = uniconv.convertUni(value)
            except uniconv.UnconvertableError:
                raise ArbUnitValueError(key, value)
            except uniconv.NotSearchableError:
                # value probably is already a number
                pass
            self.arbUnitDict[key] = value

    def __setitem__(self, key, value):
        self.setitems(((key, value),))

    def __getitem__(self, key):
        return self.arbUnitDict[key]

    def __delitem__(self, key):
        del self.arbUnitDict[key]

    def keys(self):
        return self.arbUnitDict.keys()

class _FoodBase(object):
    """ Base object for foods.

    Both regular Foods and Recipes are derived from this base class. """
    
    def __init__(self, traits, arbUnits, origFile):
        self._arbUnits = _ArbUnits(self, arbUnits)
        self._traits = _Traits(self, traits)
        self._origFile = origFile

    def __str__(self):
        import pprint
        rep = 'Traits:\n%s\n' % pprint.pformat(self.traits)
        rep = rep + 'Gram weight: %s\n' % self.grams
        rep = rep + 'nuts:\n%s\n' % pprint.pformat(self.nuts)
        rep = rep + 'Units:\n%s\n' % pprint.pformat(self.units)
        return rep

    def _getOrigFile(self):
        """The file this food originally came from."""
        return self._origFile

    def _setOrigFile(self, filename):
        self._origFile = filename

    def _delOrigFile(self):
        del self._origFile

    origFile = property(fget = _getOrigFile,
            fset = _setOrigFile,
            fdel = _delOrigFile,
            doc = _getOrigFile.__doc__)

    def _getGrams(self):
        return self._grams

    grams = property(fget = _getGrams,
            doc = "Weight of this food, considering quantity and units.")

    def _getArbUnits(self):
        return self._arbUnits

    arbUnits = property(fget = _getArbUnits,
            doc = """ Mapping of arbitrary units in this food.

            Keys are unit names; values are amounts. """)

    def _getTraits(self):
        """ Traits of this food.

        Traits are set using unicode, even for floating-point
        traits. """
        return self._traits

    traits = property(fget = _getTraits, doc = _getTraits.__doc__ )

    def _getQtyMinusRefuse(self):
        """ Gets the quantity of this food, reduced by its percent refuse. """
        import uniconv

        pctRefuse = uniconv.convertUni(self.traits['pctRefuse'])
        oldQty = uniconv.convertUni(self.traits['qty'])
        try:
            newQty = oldQty - pctRefuse / 100 * oldQty
        except TypeError:
            # pctRefuse is None
            newQty = ''
        return str(newQty)

    qtyMinusRefuse = property(fget = _getQtyMinusRefuse,
            doc = """ Quantity of food, reduced by percent refuse. """)

    def _getHash(self):
        """ Hash of this food.

        The hash is only of the food's traits; nuts and arbUnits are
        not considered when calculating the hash. """
        
        # hashlib new in python-2.5
        try:
            import hashlib
            hasher = hashlib.sha1
        except ImportError:
            import sha 
            hasher = sha.sha

        fingerPrint = ''
        for traitName in TRAITNAMES:
            fingerPrint += self._traits[traitName].encode('utf-8')
        return hasher(fingerPrint).digest()

    hash = property(fget = _getHash, doc = _getHash.__doc__)

    def getQtyByNut(self, sNutName, nutAmt, options):
        """ Return qty needed for amount of nutName to equal nutAmt.

        options are set using global constants IGNORECASE and EXACTMATCH.
        getQtyByNut looks up nutName using these options. For no options, 
        use 0.
        
        nutAmt is a Unicode. """

        import uniconv

        nutNames = [ key[0] for key in self.nuts.keys() ]
        matches = _listSearch(nutNames, sNutName, options)
        if len(matches) != 1:
            raise NutrientSearchError(self, sNutName, matches, nutAmt)
        nutName = matches[0]
        nutValue = self.nuts[[nutName]]
        try:
            nutAmt = uniconv.convertUni(nutAmt)
        except uniconv.UnconvertableError:
            raise BadAmtError(self, sNutName, matches, nutAmt)

        foodQty = ((uniconv.convertUni(self.traits['qty']) * nutAmt) 
                / nutValue)
        return str(round(foodQty,2))

    def setTraits(self, newTraits):
        """ Set traits en masse. """
        try:
            newTraits = newTraits.items()
        except AttributeError:
            pass

        for key, value in newTraits:
            self.traits[key] = value

    def setUnit(self, pattern, options):
        """ Set unit of this food to the unit matching pattern. """
        matches = _listSearch(self.units, pattern, options)
        if len(matches) != 1:
            raise SetUnitError(self, pattern, matches) 
        self.traits['unit'] = matches[0]

    def _mutant(self):
        """ Returns copy of food, with Comment field changed.

        Handy for adding foods to a cabinet without adding dupes. 
        Returns 2-tuple, with first element being the mutated food,
        and the second element being an integer with the copy number. """

        import re
        newFood = self.copy()
        regex = re.compile(r'\(Copy ([0-9]+)\) (.*)')
        match = regex.match(newFood.traits['comment'])
        if match:
            mutex = int(match.group(1)) + 1
            newFood.traits['comment'] = ('(Copy %u) %s' 
                    % (mutex, match.group(2)))
        else:
            mutex = 1
            newFood.traits['comment'] = ('(Copy 1) %s' 
                    % newFood.traits['comment'])
        return (newFood, mutex)

    mutant = property(fget=_mutant, doc=_mutant.__doc__)

class Recipe(_FoodBase):
    """ Defines a recipe. """
    
    def __init__(self, traits, arbUnits, origFile, ingredients, rTraits): 
        """ Set up a new recipe. 
        
        traits: a mapping with the recipe's traits. Each key may be
        selected either from the keys in TRAITSBYNO (integers) or
        the keys in TRAITSBYNAME (strings). The values are always,
        always Unicode objects--even for seemingly "numeric" traits
        such as 'qty' or 'pctRefuse'. In this manner, fractions and
        compound numbers may be used for numeric traits. If you do
        not set a trait in this mapping, it will be initialized to a
        zero-length Unicode. EXCEPTION: you must always set a
        non-zero-length value for the qty and unit traits; if you do
        not set these traits, NoQtyError or NoUnitError is raised.

        arbUnits: a mapping with the recipe's arbitrary units. The
        keys are Unicode objects with the name of the arbitrary
        unit; the values are numbers with the weight of the
        arbitrary unit, in grams.

        origFile: a Unicode which will indicate where this recipe
        originally came from. For example, all the recipes in a Txt
        cabinet originally came from a different cabinet. The
        filename of the original cabinet is indicated here. 

        ingredients: tuple of foods that comprise the ingredients
        for this recipe. Each value may be a Food or Recipe instance
        for the ingredient. Alternatively, the value may be a result
        from a Food or Recipe's data() method.

        rTraits: a mapping with the recipe's recipe traits. Each key
        may be selected either from the keys in RTRAITSBYNO
        (integers) or the keys in RTRAITSBYNAME (strings). The values
        are always, always Unicode objects--even for seemingly
        "numeric" traits such as 'yield' or 'servings'. In this
        manner, fractions and compound numbers may be used for
        numeric traits. If you do not set a trait in this mapping,
        it will be initialized to a zero-length Unicode. Unlike the
        traits object, you do not need to set any of these; to not
        set any recipe traits, pass an empty dict for the rTraits
        parameter. If you do not set a value for the 'yield' recipe
        trait, Pantry will guess the recipe yield based on the mass
        of all the ingredients.

        If a value is set for the servings rTrait, then Pantry will
        make an appropriate arbitrary unit for servings.

        """
        self._rTraits = _RTraits(self, rTraits)
        self._ingredients = []
        for ingredient in ingredients:
            try:
                self._ingredients.append(Food(**ingredient))
            except TypeError:
                try:
                    self._ingredients.append(Recipe(**ingredient))
                except TypeError:
                    self._ingredients.append(ingredient)
        super(Recipe, self).__init__(traits, arbUnits, origFile)

    def __str__(self):
        import pprint
        rep = _FoodBase.__str__(self)
        rep += 'rTraits:\n%s\n' % pprint.pformat(self.rTraits)
        rep += 'INGREDIENTS:\n'
        for ingredient in self._ingredients:
            rep += str(ingredient)
        return rep

    def _getTotalGrams(self):
        """ Total grams yielded by one batch of this recipe.

        Not to be confused with grams attribute, which is the total
        grams of a single portion of the recipe, as determined by the
        qty and unit attributes. """
        from uniconv import convertUni
        
        yieldNo = convertUni(self.rTraits['yield'])
        if yieldNo:
            return yieldNo
        return sum([food.grams for food in self._ingredients])

    totalGrams = property(fget=_getTotalGrams, doc=_getTotalGrams.__doc__)

    def _getNuts(self):
        """ Mapping instance of nutrients in this food.

        Keys are (nutName, nutUnits); values are nutrient amounts. """

        grams = self.grams
        totalGrams = self.totalGrams
        return dict([(nutId, nutValue / totalGrams * grams)
            for nutId, nutValue in nutSum(self.ingredients).items()])

    nuts = property(fget=_getNuts, doc=_getNuts.__doc__)

    def _getData(self):
        """ Dict with data that may be used to reconstruct this recipe.

        e.g. to copy the recipe, Recipe(**recipe.data). Every value
        in the dict has only Python primatives, making it easy to
        pickle.
        """
        return {'traits':self._traits.traitdict.copy(),
                'arbUnits':self._arbUnits.arbUnitDict.copy(),
                'origFile':self._origFile,
                'ingredients':[ food.data for food in self._ingredients ],
                'rTraits':self._rTraits.traitdict}

    data = property(fget = _getData, doc = _getData.__doc__)

    def _getIngredients(self):
        """ List of foods that comprise this recipe's ingredients. """
        return self._ingredients

    ingredients = property(fget=_getIngredients, doc=_getIngredients.__doc__)

    def _getRTraits(self):
        """ Traits specific to recipes. """
        return self._rTraits

    rTraits = property(fget=_getRTraits, doc=_getRTraits.__doc__)

    def _getUnits(self):
        """ Units available for this food.

        Absolute units "mask" arbitrary units--that is,
        if there is an absolute unit and an arbitrary unit
        with the same description, this attribute contains only 
        the absolute unit. 
        
        Units for foods also include a 'serving' unit if
        the 'servings' rTrait is set."""
        import uniconv
        result = dict(self._arbUnits.items() + ABSOLUTES.items())
        servings = uniconv.convertUni(self._rTraits['servings'])
        if servings:
            result['serving'] = self.totalGrams / servings
        return result

    units = property(fget = _getUnits, doc = _getUnits.__doc__ )

    def copy(self):
        """ Return a deep copy of this recipe. """
        return Recipe(**self.data)

class Food(_FoodBase):
    """ Defines a food. """

    def __init__(self, traits, arbUnits, origFile, nuts):
        """ Set up a new food. 

        traits: a mapping with the food's traits. Each key may be
        selected either from the keys in TRAITSBYNO (integers) or
        the keys in TRAITSBYNAME (strings). The values are always,
        always Unicode objects--even for seemingly "numeric" traits
        such as 'qty' or 'pctRefuse'. In this manner, fractions and
        compound numbers may be used for numeric traits. If you do
        not set a trait in this mapping, it will be initialized to a
        zero-length Unicode. EXCEPTION: you must always set a
        non-zero-length value for the qty and unit traits; if you do
        not set these traits, NoQtyError or NoUnitError is raised.

        arbUnits: a mapping with the food's arbitrary units. The
        keys are Unicode objects with the name of the arbitrary
        unit; the values are numbers with the weight of the
        arbitrary unit, in grams.

        origFile: a Unicode which will indicate where this food
        originally came from. For example, all the foods in a Txt
        cabinet originally came from a different cabinet. The
        filename of the original cabinet is indicated here. 

        nuts: mapping with the food's nutrients. Each key may be
        selected from the keys in NUTSBYNO (integers) or may be a
        tuple.  Which you use is of critical importance. If the key
        is an integer, Pantry assumes that the value has already
        been converted to amount of nutrient per 100g (which is how
        Pantry stores nutrient values internally.) 
        
        If the key is a tuple, then the first value of the tuple
        must be a string, which must correspond to one of the
        nutrients in NUTSBYNAME. The second value of the tuple is
        optional. If the second value is '%', Pantry will convert
        the nutrient value depending upon the nutrient's %DV, if it
        has a percent DV. If the second value is anything else, then
        Pantry will check the units to make sure they are
        permissible (e.g. that the units for Protein are 'g', not
        'mg'). If the second value is not set, Pantry will do no
        checking. In any case, if the key is a tuple, then Pantry
        will convert it to nutrients per 100 g for you; the result
        of the calculation will depend on the values for the 'unit'
        and 'qty' traits, so be sure these are set correctly. 

        If the key is a tuple, the value must be a Unicode; if the
        key is an integer, the value must be an integer or float.

        Some example key/value pairs:
            {203: 20} Pantry will set the new food's nutrients per 100g to
            20.

            {('Protein',): u'20'} Pantry will convert 20 to nutrients
            per 100 g.

            {('Protein', '%'): u'20'} Pantry will convert 20 to the
            appropriate nutrients per 100g, using the nutrient's
            %DV.

            {('Protein', 'g'): u'20'} Pantry will make sure that 'g' is
            the correct unit to use for this nutrient and, if it
            is, will convert 20 to nutrients per 100g.

            {(203,): 20} Pantry will crash with some sort of
            error. The key is invalid.

            {'Protein':20} Pantry will crash with some sort of
            error. The key is invalid.

            {('Protein', 'g'):20} Pantry will crash with some sort
            of error. The value is a integer, but because the key is
            a tuple, the value must be a Unicode.
        
        It is permissible to construct a food with no nutrients,
        which is sometimes desirable to save time and memory; in
        this case, pass an empty dictionary for nuts.  """

        _FoodBase.__init__(self, traits, arbUnits, origFile)
        self._nuts = _Nuts(self, nuts = nuts)

    def _getData(self):
        """ Dict with data that may be used to reconstruct this food.

        e.g. to copy the food, you can do Food(**food.data).
        """

        return {'traits':self._traits.traitdict.copy(), 
                'arbUnits':self._arbUnits.arbUnitDict.copy(), 
                'origFile':self._origFile, 
                'nuts':self._nuts.nutdict.copy()}

    data = property(fget = _getData, doc = _getData.__doc__)

    def copy(self):
        """ Return a deep copy of this food. """
        return Food(**self.data)

    def _getUnits(self):
        """ Units available for this food.

        Absolute units "mask" arbitrary units--that is,
        if there is an absolute unit and an arbitrary unit
        with the same description, this attribute contains only 
        the absolute unit. """
        return dict(self._arbUnits.items() + ABSOLUTES.items())

    units = property(fget = _getUnits, doc = _getUnits.__doc__ )

    def _getNuts(self):
        """ Mapping of nutrients in this food.

        Keys are (nutName, nutUnits); values are integers or floats
        with the amount of the nutrient. """
        return self._nuts

    nuts = property(fget = _getNuts, doc = _getNuts.__doc__)

class CmpFoods(object):
    """ Compare two foods for sorting purposes. 
    
    Mimics Python's cmp function; that is, returns -1 if x is less
    than y, 0 if x == y, or 1 if x is greater than y. """

    def __init__(self, traits, sortOrders):
        """ Set up a food comparison object.

        traits is a Unicode with the first letter of each trait that
        will be used as a comparison key. For example, to compare
        foods by name, and then by group if the names are equal,
        pass u'ng' for traits. Comparisons are ascending, unless the
        trait name is preceded by u'-', in which case they are
        descending. So, to compare by names in ascending order and
        then by group in descending order, pass u'n-g' for traits.

        sortOrders is a mapping. The keys are the names of the
        traits; the values are tuples specifying what order the
        traits are to be sorted in. For example, to specify that the
        meal trait is to be sorted in the order 'Breakfast, Lunch,
        Dinner', pass {'meal':('Breakfast', 'Lunch', 'Dinner')}
        for sortOrders. If a sort order is specified for a
        particular trait, and one food has a trait that is in the
        sort order and another food does not have a trait that is in
        the sort order, the food whose trait is in the sort order is
        sorted superior to the one whose trait is not in the sort
        order.
        """
        self.traits = traits
        self.sortOrders = sortOrders

    def __call__(self, x, y):

        for trait in self.traits:
            invert = False
            if trait.startswith('-'):
                invert = True
                trait = trait.lstrip('-')

            xTrait = x.traits[trait]
            yTrait = y.traits[trait]
            if trait in self.sortOrders:
                sortOrder = self.sortOrders[trait]
                if xTrait in sortOrder and yTrait in sortOrder:
                    result = cmp(sortOrder.index(xTrait),
                            sortOrder.index(yTrait))
                elif xTrait not in sortOrder and yTrait not in sortOrder:
                    result = cmp(xTrait, yTrait)
                elif xTrait in sortOrder and yTrait not in sortOrder:
                    result = 1
                elif xTrait not in sortOrder and yTrait in sortOrder:
                    result = -1
            else:
                result = cmp(xTrait, yTrait)
            
            if invert:
                result = -result
            if result != 0:
                return result
        return 0

class NutInfo(object):
    """ Holds information about a nutrient.

    Attributes:
        no 
        name 
        units  
        dv
    """
    def __init__(self, no, name, units, dv):
        self.no = no
        self.name = name
        self.units = units
        self.dv = dv

def _listSearch(sList, sStr, options):
    """ Returns list of items from sList that match sStr.
    
    Options are set using the global constants IGNORECASE
    and EXACTMATCH."""

    if options & EXACTMATCH:
        return [ item for item in sList if item == sStr ]
    else:
        import re
        flag = 0
        if options & IGNORECASE:
            flag = re.IGNORECASE
        sRE = re.compile(sStr, flag)

        return [ item for item in sList if sRE.search(item) ]

def nutSum(foods):
    """ Sums nutrients.

    foods is a tuple of instances having a nuts attribute. nutSum
    sums these and returns a single mapping with the sum. """

    result = {}
    for food in foods:
        for nutID, nutValue in food.nuts.items():
            try:
                result[nutID] += nutValue
            except KeyError:
                result[nutID] = nutValue
    return result

### EXCEPTIONS
class Error(Exception):
    """ Base class for exceptions in this module."""
    pass

class SetUnitError(Error):
    """ Could not set unit to unit requested. """
    def __init__(self, food, pattern, matches):
        self.msg  = [u'Failed to set units using pattern %s' % pattern]
        self.msg.append('Matches:')
        if len(matches) > 0:
            self.msg.append('\n'.join(matches))
        else:
            self.msg.append('<none>')

class QtyByNutError(Error):
    """ getQtyByNut errors. """
    def __init__(self, food, pattern, matches, nutAmt):
        self.msg = [u'Failed to find qty of food per amount nutrient.']

class NutrientSearchError(QtyByNutError):
    """ getQtyByNut search returned != 1 match. """
    def __init__(self, food, pattern, matches, nutAmt):
        QtyByNutError.__init__(self, food, pattern, matches, nutAmt)
        self.msg.append(u'Pattern %s did not yield exactly one match.'
                % pattern)
        self.msg.append(u'Matches:')
        if len(matches) > 1:
            self.msg.append(u'\n'.join(matches))
        else:
            self.msg.append(u'<none>')

class BadAmtError(QtyByNutError):
    """ getQtyByNut failed because amount would not convert. """
    def __init__(self, food, pattern, matches, nutAmt):
        QtyByNutError.__init__(self, food, pattern, matches, nutAmt)
        self.msg.append('Value %s is not a number.' % nutAmt)

class SetNutError(Error):
    """ Errors that occur in Nuts.__setitem__. """
    def __init__(self, key, value):
        self.key = key
        self.value = value
        self.msg = [u'Failed to set nutrient %s' % key[0]]

class NutIDError(SetNutError):
    """ Nutrient ID not found.

    Raised if nutrient number is not found, or if nutrient name is
    not found if nutrient is being set by name. Do not raise if nutrient
    number or name is found, but units do not match. """
    def __init__(self, key, value):
        SetNutError.__init__(self, key, value)
        self.msg.append('Nutrient name is invalid.')

class NutValueError(SetNutError):
    """ Nutrient value is erroneous.

    e.g. nutrient value will not convert to a float. """

    def __init__(self, key, value):
        SetNutError.__init__(self, key, value)
        self.msg.append('Nutrient value %s is not a number.' % value)

class NoPercentageError(SetNutError):
    """ Units were set to percentage, but there's no %DV for this nut. """
    
    def __init__(self, key, value):
        SetNutError.__init__(self, key, value)
        self.msg.append('No percent daily value available.')

class NutUnitMismatchError(SetNutError):
    """ Units given do not match the units for this nutrient. """
    
    def __init__(self, key, value, validUnit):
        SetNutError.__init__(self, key, value)
        self.msg.append('Given unit: %s Valid unit: %s' % (key[1], validUnit))

class SetTraitError(Error):
    """ Errors that occur in Traits.__setitem__. """
    def __init__(self, key, value):
        self.key = key
        self.value = value
        self.msg = [u'Failed to set trait %s' % key]

class NoAvailUnitError(SetTraitError):
    """ No available unit for given unit. """
    def __init__(self, key, value):
        SetTraitError.__init__(self, key, value)
        self.msg.append(u'Unit not available: %s' % value)

class TraitValueError(SetTraitError):
    """ Trait value will not convert to float. """
    
    def __init__(self, key, value):
        SetTraitError.__init__(self, key, value)
        self.msg.append(u'Value %s is not a number' % value)

class TraitInitError(Error):
    """ Problems initializing a food's traits. """
    def __init__(self):
        self.msg = [u'Could not initialize traits:']

class NoQtyError(TraitInitError):
    """ Traits do not include a qty trait. """
    def __init__(self):
        TraitInitError.__init__(self)
        self.msg.append(u'qty trait not set.')

class NoUnitError(TraitInitError):
    """ Traits do not include a unit trait. """
    def __init__(self):
        TraitInitError.__init__(self)
        self.msg.append(u'unit trait not set.')

class SetArbUnitError(Error):
    """ Errors that occur in ArbUnits.__setitem__. """
    def __init__(self, key, value):
        self.key = key
        self.value = value
        self.msg = [u'Failed to set unit %s:' % key]

class ArbUnitValueError(SetArbUnitError):
    """ Arb unit will not convert to a float. """

    def __init__(self, key, value):
        SetArbUnitError.__init__(self, key, value)
        self.msg.append('Value %s is not a number' % value)

### CONSTANTS
ALLNUTS = (
    (203, 'Protein', 'g', 50),
    (204, 'Total Fat', 'g', 65),
    (205, 'Total Carbohydrate', 'g', 300),
    (207, 'Ash', 'g', None),
    (208, 'Calories', 'kcal', None),
    (209, 'Starch', 'g', None),
    (210, 'Sucrose', 'g', None),
    (211, 'Glucose (dextrose)', 'g', None),
    (212, 'Fructose', 'g', None),
    (213, 'Lactose', 'g', None),
    (214, 'Maltose', 'g', None),
    (221, 'Alcohol, ethyl', 'g', None),
    (255, 'Water', 'g', None),
    (257, 'Adjusted Protein', 'g', None),
    (262, 'Caffeine', 'mg', None),
    (263, 'Theobromine', 'mg', None),
    (268, 'Energy', 'kj', None),
    (269, 'Sugars', 'g', None),
    (287, 'Galactose', 'g', None),
    (291, 'Dietary Fiber', 'g', 25),
    (301, 'Calcium', 'mg', 1000),
    (303, 'Iron', 'mg', 18),
    (304, 'Magnesium', 'mg', 400),
    (305, 'Phosphorus', 'mg', 1000),
    (306, 'Potassium', 'mg', 3500),
    (307, 'Sodium', 'mg', 2400),
    (309, 'Zinc', 'mg', 15),
    (312, 'Copper', 'mg', 2),
    (313, 'Fluoride', 'mcg', None),
    (315, 'Manganese', 'mg', None),
    (317, 'Selenium', 'mcg', None),
    (318, 'Vitamin A', 'IU', 5000),
    (319, 'Retinol', 'mcg', None),
    (320, 'Vitamin A, RAE', 'mcg_RAE', None),
    (321, 'Carotene, beta', 'mcg', None),
    (322, 'Carotene, alpha', 'mcg', None),
    (323, 'Vitamin E', 'mg', None),
    (324, 'Vitamin D', 'IU', 400),
    (334, 'Cryptoxanthin, beta', 'mcg', None),
    (337, 'Lycopene', 'mcg', None),
    (338, 'Lutein + zeaxanthin', 'mcg', None),
    (341, 'Tocopherol, beta', 'mg', None),
    (342, 'Tocopherol, gamma', 'mg', None),
    (343, 'Tocopherol, delta', 'mg', None),
    (401, 'Vitamin C', 'mg', 60),
    (404, 'Thiamin', 'mg', 1.5),
    (405, 'Riboflavin', 'mg', 1.7),
    (406, 'Niacin', 'mg', 20),
    (410, 'Pantothenic acid', 'mg', 10),
    (415, 'Vitamin B-6', 'mg', 2),
    (417, 'Folate, total', 'mcg', None),
    (418, 'Vitamin B-12', 'mcg', 6),
    (421, 'Choline, total', 'mg', None),
    (430, 'Vitamin K (phylloquinone)', 'mcg', None),
    (431, 'Folic acid', 'mcg', 400),
    (432, 'Folate, food', 'mcg', None),
    (435, 'Folate, DFE', 'mcg_DFE', None),
    (454, 'Betaine', 'mg', None),
    (501, 'Tryptophan', 'g', None),
    (502, 'Threonine', 'g', None),
    (503, 'Isoleucine', 'g', None),
    (504, 'Leucine', 'g', None),
    (505, 'Lysine', 'g', None),
    (506, 'Methionine', 'g', None),
    (507, 'Cystine', 'g', None),
    (508, 'Phenylalanine', 'g', None),
    (509, 'Tyrosine', 'g', None),
    (510, 'Valine', 'g', None),
    (511, 'Arginine', 'g', None),
    (512, 'Histidine', 'g', None),
    (513, 'Alanine', 'g', None),
    (514, 'Aspartic acid', 'g', None),
    (515, 'Glutamic acid', 'g', None),
    (516, 'Glycine', 'g', None),
    (517, 'Proline', 'g', None),
    (518, 'Serine', 'g', None),
    (521, 'Hydroxyproline', 'g', None),
    (573, 'Vitamin E, added', 'mg', None),
    (578, 'Vitamin B-12, added', 'mcg', None),
    (601, 'Cholesterol', 'mg', 300),
    (605, 'Trans Fat', 'g', None),
    (606, 'Saturated Fat', 'g', 20),
    (607, '4:0', 'g', None),
    (608, '6:0', 'g', None),
    (609, '8:0', 'g', None),
    (610, '10:0', 'g', None),
    (611, '12:0', 'g', None),
    (612, '14:0', 'g', None),
    (613, '16:0', 'g', None),
    (614, '18:0', 'g', None),
    (615, '20:0', 'g', None),
    (617, '18:1 undifferentiated', 'g', None),
    (618, '18:2 undifferentiated', 'g', None),
    (619, '18:3 undifferentiated', 'g', None),
    (620, '20:4 undifferentiated', 'g', None),
    (621, '22:6 n-3', 'g', None),
    (624, '22:0', 'g', None),
    (625, '14:1', 'g', None),
    (626, '16:1 undifferentiated', 'g', None),
    (627, '18:4', 'g', None),
    (628, '20:1', 'g', None),
    (629, '20:5 n-3', 'g', None),
    (630, '22:1 undifferentiated', 'g', None),
    (631, '22:5 n-3', 'g', None),
    (636, 'Phytosterols', 'mg', None),
    (638, 'Stigmasterol', 'mg', None),
    (639, 'Campesterol', 'mg', None),
    (641, 'Beta-sitosterol', 'mg', None),
    (645, 'Monounsaturated Fat', 'g', None),
    (646, 'Polyunsaturated Fat', 'g', None),
    (652, '15:0', 'g', None),
    (653, '17:0', 'g', None),
    (654, '24:0', 'g', None),
    (662, '16:1 t', 'g', None),
    (663, '18:1 t', 'g', None),
    (664, '22:1 t', 'g', None),
    (665, '18:2 t not further defined', 'g', None),
    (666, '18:2 i', 'g', None),
    (669, '18:2 t,t', 'g', None),
    (670, '18:2 CLAs', 'g', None),
    (671, '24:1 c', 'g', None),
    (672, '20:2 n-6 c,c', 'g', None),
    (673, '16:1 c', 'g', None),
    (674, '18:1 c', 'g', None),
    (675, '18:2 n-6 c,c', 'g', None),
    (676, '22:1 c', 'g', None),
    (685, '18:3 n-6 c,c,c', 'g', None),
    (687, '17:1', 'g', None),
    (689, '20:3 undifferentiated', 'g', None),
    (693, 'Fatty acids, total trans-monoenoic', 'g', None),
    (695, 'Fatty acids, total trans-polyenoic', 'g', None),
    (696, '13:0', 'g', None),
    (697, '15:1', 'g', None),
    (851, '18:3 n-3 c,c,c', 'g', None),
    (852, '20:3 n-3', 'g', None),
    (853, '20:3 n-6', 'g', None),
    (855, '20:4 n-6', 'g', None),
    (856, '18:3i', 'g', None),
    (857, '21:5', 'g', None),
    (858, '22:4', 'g', None)
)

ALLNUTINFOS = [ NutInfo(_no, _name, _units, _dv) for _no, _name, _units, _dv
        in ALLNUTS ]
NUTSBYNO = dict([(_nut.no, _nut) for _nut in ALLNUTINFOS])
NUTSBYNAME = dict([(_nut.name, _nut) for _nut in ALLNUTINFOS])

IGNORECASE = 1
EXACTMATCH = 2

TRAITNAMES = ('name', 'group', 'refDesc', 'pctRefuse',
        'unit', 'date', 'meal', 'qty', 'comment', 'order')
FLOATTRAITS = ('qty', 'pctRefuse')
TRAITSBYNO = dict(zip(range(len(TRAITNAMES)), TRAITNAMES))
TRAITSBYNAME = dict(zip(TRAITNAMES, range(len(TRAITNAMES))))
RTRAITNAMES = ('yield', 'yieldDesc', 'servings', 'directions')
RFLOATTRAITS = ('yield', 'servings')
RTRAITSBYNO = dict(zip(range(len(RTRAITNAMES)), RTRAITNAMES))
RTRAITSBYNAME = dict(zip(RTRAITNAMES, range(len(RTRAITNAMES))))
ABSOLUTES = { u'g':1.0, u'oz':28.35, u'lb':453.59 }
